# ==========
# Basic model infrastructure
# Do not run this file; import it and run your file.
# ==========

import os, sys, warnings, asyncio, time
import gettext
from random import shuffle, choice
from typing import Callable
from numpy import random
import pandas
#from memory_profiler import profile

from helipad.visualize import BaseVisualization, Charts, TimeSeries
from helipad.helpers import *
from helipad.param import Params, Shocks
from helipad.data import Data
from helipad.agent import Agent, baseAgent

class Helipad:
	runInit = True #for multiple inheritance. Has to be a static property

	def __init__(self, locale:str='en'):
		#Have to do this first so that i18n is available early.
		#Put it in an obscure variable and then use helpers.ï() so we don't conflict with the REPL console.
		if not hasattr(self, 'breed'):
			import builtins
			builtins.__dict__['helipad_gettext'] = gettext.translation('helipad', localedir=os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))+'/locales', languages=[locale]).gettext

		self.data = Data(self)
		self.params = Params(self)
		self.shocks = Shocks(self)
		self.events = Events()
		self.hooks = Hooks()
		self.primitives = Primitives(self)
		self.goods = Goods(self)				#List of goods

		self.name = ''
		self.agents = {}
		self.patches = []
		self.stages = 1
		self.order = 'linear'
		self.hasModel = False		#Have we initialized?
		self.timer = False
		self.visual = None
		self.cpanel = None

		self.t = None
		self.running = False
		self._cut = False

		#Default parameters
		self.primitives.add('agent', Agent, dflt=50, low=1, high=100)

		#Decorators
		def repdec(name, fn, kwargs): self.data.addReporter(name, fn, **kwargs)
		def hookdec(name, fn, kwargs): self.hooks.add(name, fn, **kwargs)
		def buttondec(name, fn, kwargs): self.addButton(name, fn, **kwargs)
		def eventdec(name, fn, kwargs): self.events.add(name, fn, **kwargs)
		self.reporter = self.genDecorator(repdec)
		self.hook = self.genDecorator(hookdec)
		self.button = self.genDecorator(buttondec)
		self.event = self.genDecorator(eventdec)

		#A few things that only make sense to do if it's the topmost model
		if not hasattr(self, 'breed'):

			#Privileged parameters
			#Toggle the progress bar between determinate and indeterminate when stopafter gets changed
			def switchPbar(model, name, val):
				if not model.hasModel or not model.cpanel or not getattr(model.cpanel, 'progress', False): return
				if not val: model.cpanel.progress.determinate(False)
				else:
					model.cpanel.progress.determinate(True)
					model.cpanel.progress.update(model.t/val)
			self.params.add('stopafter', ï('Stop on period'), 'checkentry', False, runtime=True, config=True, entryType='int', callback=switchPbar)
			self.params.add('csv', ï('CSV?'), 'checkentry', False, runtime=True, config=True)
			self.params.add('refresh', ï('Refresh Every __ Periods'), 'slider', 20, opts=[1, 2, 5, 10, 20, 50, 100, 200, 500, 1000], runtime=True, config=True)
			self.params['shocks'] = self.shocks

			#Check for updates
			from helipad.__init__ import __version__
			import xmlrpc.client, ssl

			#Is v1 bigger than v2?
			#There's a `packaging` function to do this, but let's not bloat our dependencies.
			def vcompare(v1, v2):
				(v1, v2) = (v1.split('.'), (v2.split('.')))

				#Pad major releases with zeroes to make comparable
				maxl = max([len(v1), len(v2)])
				for v in [v1,v2]:
					if len(v)<maxl: v += [0 for i in range(maxl-len(v))]

				for k,i in enumerate(v1):
					if i > v2[k]: return True
					elif i < v2[k]: return False
				return False

			try:
				pypi = xmlrpc.client.ServerProxy('https://pypi.org/pypi', context=ssl._create_unverified_context())
				available = pypi.package_releases('helipad')
				if vcompare(available[0], __version__):
					print(ï('A Helipad update is available! Use `pip install -U helipad` to upgrade to version {}.').format(available[0]))
			except: pass #Fail silently if we're not online

	def __repr__(self):
		if self.name: return f'<Helipad: {self.name}>'
		else: return '<Helipad object>'

	def addButton(self, text, func, desc=None):
		self.shocks.add(text, None, func, 'button', True, desc)

	#Get or set a parameter, depending on whether there are two or three arguments
	def param(self, param, val=None):
		item = param[2] if isinstance(param, tuple) and len(param)>2 else None
		param = self.params[param[0]] if isinstance(param, tuple) else self.params[param]

		if val is not None: param.set(val, item)
		else: return param.get(item)

	def addBreed(self, name, color, prim=None):
		if prim is None:
			if len(self.primitives) == 1: prim = next(iter(self.primitives.keys()))
			else: raise KeyError(ï('Breed must specify which primitive it belongs to.'))
		return self.primitives[prim].breeds.add(name, color)

	#Returns the value of the last function in the list
	def doHooks(self, place, args):
		#Take a list of hooks; go until we get a response
		if isinstance(place, list):
			for f in place:
				r = self.doHooks(f, args)
				if r is not None: return r
			return None

		if not place in self.hooks: return None
		for f in self.hooks[place]: r = f(*args)
		return r

	def useVisual(self, viz):
		if hasattr(self, 'breed'):
			warnings.warn(ï('Visualizations can only be registered on the top-level model.'), None, 2)
			return #Doesn't matter if it's not the top-level model

		if viz is not None and not issubclass(viz, BaseVisualization):
			raise RuntimeError(ï('Visualization class must inherit from BaseVisualization.'))

		self.visual = viz(self) if viz is not None else None
		return self.visual

	#Get ready to actually run the model
	def setup(self):
		if self.hasModel: self.terminate()
		self.doHooks('modelPreSetup', [self])
		self.t = 0

		#Blank breeds for any primitives not otherwise specified
		for p in self.primitives.values():
			if not p.breeds: p.breeds.add('', '#000000')

		#SERIES AND REPORTERS
		#Breeds and goods should already be registered at this point

		self.data.reset()
		for e in self.events.values(): e.reset()
		defPrim = 'agent' if 'agent' in self.primitives else next(iter(self.primitives))

		def pReporter(param, item=None):
			def reporter(model): return param.get(item)
			return reporter

		#Add reporters for parameters
		for n,p in self.params.items():
			if p.type == 'hidden' or getattr(p, 'config', False): continue	#Skip hidden and config parameters
			if p.per is None: self.data.addReporter(n, pReporter(p))
			elif p.per == 'good':
				for good in p.pKeys:
					self.data.addReporter(n+'-'+good, pReporter(p, good))
			elif p.per == 'breed':
				for breed in p.pKeys:
					self.data.addReporter(p.prim+'-'+n+'-'+breed, pReporter(p, breed))

		if self.goods.money is not None:
			self.data.addReporter('M0', self.data.agentReporter('stocks', 'all', good=self.goods.money, stat='sum'))
			if self.visual is not None and isinstance(self.visual, TimeSeries) and 'money' in self.visual.plots:
				self.visual.plots['money'].addSeries('M0', ï('Monetary Base'), self.goods[self.goods.money].color)

		#Unconditional variables to report
		# self.data.addReporter('utility', self.data.agentReporter('utils', defPrim))

		#Per-breed and per-good series and reporters
		#Don't put lambda functions in here, or the variable pairs will be reported the same, for some reason.
		for breed, b in self.primitives[defPrim].breeds.items():
			self.data.addReporter('utility-'+breed, self.data.agentReporter('utils', defPrim, breed=breed))
			if self.visual is not None and self.visual.__class__.__name__=='TimeSeries':
				self.visual.plots['utility'].addSeries('utility-'+breed, breed.title()+' '+ï('Utility'), b.color)

		if len(self.goods) >= 2:
			for good, g in self.goods.nonmonetary.items():
				self.data.addReporter('demand-'+good, self.data.agentReporter('currentDemand', 'all', good=good, stat='sum'))
				if self.visual is not None:
					if 'demand' in self.visual.plots: self.visual.plots['demand'].addSeries('demand-'+good, good.title()+' '+ï('Demand'), g.color)

		#Initialize agents
		self.primitives = dict(sorted(self.primitives.items(), key=lambda d: d[1].priority))				#Sort by priority
		pops = {prim: self.param('num_'+prim) for prim in self.primitives}
		for ags in self.agents.values(): ags.clear()														#Clear any surviving agents from last run
		for prim in self.primitives: self.nUpdater(pops[prim], prim, self, force=True)						#Force is so we can call nupdater before instantiating hasModel

		#Start progress bar
		#Put this here and not in .start() because it'll flash on unpause otherwise
		if self.cpanel:
			st = self.param('stopafter')
			self.cpanel.progress.determinate(st and isinstance(st, int))

		for param in self.params.values():
			if not param.runtime: param.disable() #Disable parameters that can't be changed during runtime

		#Patch our async functions for compatibility with Spyder's event loop
		if isIpy() and not isNotebook():
			try:
				import nest_asyncio
				nest_asyncio.apply()
			except:
				raise ImportError(ï('nest_asyncio is required to run Helipad from Spyder.'))

		self.hasModel = True
		self.doHooks('modelPostSetup', [self])

	def cutStep(self): self._cut = True

	def step(self, stage:int=1) -> int:
		self.t += 1
		self.doHooks('modelPreStep', [self])

		#Reset per-period variables
		#Have to do this all at once at the beginning of the period, not when each agent steps
		for p in self.agents.values():
			for a in p: a.currentDemand = {g:0 for g in self.goods}

		self.shocks.step()

		def sortFunc(model:Helipad, stage:int, func:Callable) -> Callable:
			def sf(agent:baseAgent): return func(agent, model, stage)
			return sf

		for self.stage in range(1, self.stages+1):
			self._cut = False
			self.doHooks('modelStep', [self, self.stage])

			#Sort agents and step them
			for prim, lst in self.agents.items():
				order = self.primitives[prim].order or self.order
				if isinstance(order, list): order = order[self.stage-1]

				if order == 'random': shuffle(lst)

				#Can't do our regular doHooks() here since we want to pass the function to .sort()
				#From the user's perspective though, this doesn't matter
				#Do the more specific sorts last
				ordhooks = (self.hooks['order'] if 'order' in self.hooks else []) + (self.hooks[prim+'Order'] if prim+'Order' in self.hooks else [])
				for o in ordhooks: lst.sort(key=sortFunc(self, self.stage, o))

				#Keep the agent list constant for a given loop because modifying it while looping
				#(e.g. if an agent dies or reproduces) will screw up the looping
				agentpool = self.agents[prim].copy()

				#Matching model
				if 'match' in order:
					matchN = order.split('-')
					matchN = int(matchN[1]) if len(matchN) > 1 else 2
					matchpool = self.agents[prim].copy()
					while len(matchpool) > len(agentpool) % matchN and not self._cut:
						agents = []
						for a in range(matchN):
							agent = choice(matchpool)
							matchpool.remove(agent)
							agents.append(agent)

							#Allow selection of matches, but only on the first agent
							if a==0:
								others = self.doHooks('matchSelect', [agent, matchpool, self, self.stage])
								if others is not None:
									if (isinstance(others, baseAgent) and matchN==2): others = [others]
									if (isinstance(others, list) and len(others)==matchN-1):
										for other in others: matchpool.remove(other)
										agents += others
										break
									else: raise ValueError(ï('matchSelect did not return the correct number of agents.'))

						accept = self.doHooks('matchAccept', agents)
						if accept is not None and not accept:
							matchpool = agents + matchpool #Prepend agents to decrease the likelihood of a repeat matching
							continue
						for a in agents: agent.step(self.stage)
						self.doHooks([prim+'Match', 'match'], [agents, prim, self, self.stage])

					#Step any remainder agents
					for agent in matchpool:
						if not self._cut: agent.step(self.stage)

				#Activation model
				else:
					for a in agentpool:
						if not self._cut: a.step(self.stage)

		self.data.collect(self)
		for e in self.events.values():
			if (not e.triggered or e.repeat) and e.check(self) and self.visual is not None and not self.visual.isNull and e.name!=self.param('stopafter'):
				self.visual.event(self.t, **e.kwargs)
		self.doHooks('modelPostStep', [self])
		return self.t

	#This is split out as an async function to allow user input while running the loop in Jupyter
	async def run(self):
		if self.timer:
			begin = time.time()
		while self.running:
			t = self.step()
			st = self.param('stopafter')

			if t%self.param('refresh')==0:
				if self.timer: t2 = time.time()
				if self.cpanel and st and isinstance(st, int): self.cpanel.progress.update(t/st)

				#Update graph
				if self.visual is not None and not self.visual.isNull:
					await asyncio.sleep(0.001) #Listen for keyboard input
					data = self.data.getLast(t - self.visual.lastUpdate)

					self.visual.update(data)
					self.visual.lastUpdate = t

					self.doHooks('visualRefresh', [self, self.visual])

				elif self.cpanel:
					if isNotebook(): await asyncio.sleep(0.001) #Listen for keyboard input
					else: self.cpanel.update() #Make sure we don't hang the interface if plotless

				# Performance indicator
				if self.timer:
					newtime = time.time()
					output = ï('Period {0}: {1} periods/second ({2}% model, {3}% visuals)').format(t, round(self.param('refresh')/(newtime-begin),2), round((t2-begin)/(newtime-begin)*100,2), round((newtime-t2)/(newtime-begin)*100,2))
					if isBuffered():
						if isNotebook(): os.write(1, bytes('\r'+output, 'utf-8'))
						else: sys.stdout.write('\r'+output)
					else: print(output)
					begin = newtime

			if st:
				stop = self.events[st].triggered if isinstance(st, str) else t>=st
				if stop: self.terminate()
		if self.timer and isBuffered(): print('\r') #Newline on pause so the output doesn't interfere with the console

	#The *args allows it to be used as an Ipywidgets callback
	#@profile #To test memory usage
	def start(self, *args):

		#Start the progress bar
		if self.cpanel:
			self.cpanel.progress.start()
			self.cpanel.runButton.run()

		self.doHooks('modelStart', [self, self.hasModel])
		if not self.hasModel: self.setup()
		self.running = True

		if isNotebook(): asyncio.ensure_future(self.run()) #If Jupyter, it already has an event loop
		else: asyncio.run(self.run())	#If Tkinter, it needs an event loop

	def stop(self, *args):
		self.running = False
		if self.cpanel:
			self.cpanel.progress.stop()
			self.cpanel.runButton.pause()
		self.doHooks('modelStop', [self])

	def terminate(self, evt=False):
		if not self.hasModel: return
		self.running = False
		self.hasModel = False
		if self.visual is not None: self.visual.terminate(self) #Clean up visualizations

		remainder = int(self.t % self.param('refresh')) #For some reason this returns a float sometimes?
		if remainder > 0 and self.visual is not None and not self.visual.isNull: self.visual.update(self.data.getLast(remainder)) #Last update at the end

		if self.param('csv'): self.data.saveCSV(self.param('csv'))
		if self.cpanel:
			self.cpanel.progress.done()
			self.cpanel.runButton.terminate()

		#Re-enable parameters
		for param in self.params.values():
			if param.type=='checkentry' and param.event: continue
			if not param.runtime: param.enable()

		self.doHooks('terminate', [self, self.data.dataframe])

	#param is a string (for a global param), a name,object,item,primitive tuple (for per-breed or per-good params), or a list of such
	def paramSweep(self, param, reporters=None):
		if not self.param('stopafter'): raise RuntimeError(ï('Can\'t do a parameter sweep without the value of the \'stopafter\' parameter set.'))

		#Standardize format and get the Param objects
		if not isinstance(param, list): param = [param]
		params = {}
		for p in param:
			if not isinstance(p, tuple): p = (p,)
			pobj = self.parseParamId(p)
			params['-'.join(p)] = (p, pobj)

		#Generate the parameter space, a list of dicts
		from itertools import product
		space = [{p[0]:run[k] for k,p in enumerate(params.items())} for run in product(*[p[1].range for p in params.values()])]

		#Run the model
		alldata = []
		for i,run in enumerate(space):
			print('Run',str(i+1)+'/'+str(len(space))+':',', '.join([k+'='+('\''+v+'\'' if isinstance(v, str) else str(v)) for k,v in run.items()])+'…')
			for p in self.params.values():
				if not getattr(p, 'config', False): p.reset()

			for k,v in run.items(): params[k][1].set(v, params[k][0][2] if params[k][1].obj is not None else None)
			self.setup()
			self.start()

			if reporters is not None: data = pandas.DataFrame({k:self.data.all[k] for k in reporters})
			else: data = self.data.dataframe

			events = [Item(name=e.name, triggered=e.triggered, data=e.data) for e in self.events.values()]

			alldata.append(Item(vars=run, data=data, events=events))
		return alldata

	#Creates an unweighted and undirected network of a certain density
	def createNetwork(self, density, kind='edge', prim=None):
		if density < 0 or density > 1: raise ValueError(ï('Network density must take a value between 0 and 1.'))
		from itertools import combinations
		agents = self.allagents.values() if prim is None else self.agents[prim]
		for c in combinations(agents, 2):
			if random.randint(0,100) < density*100:
				c[0].newEdge(c[1], kind)
		return self.network(kind, prim)

	def network(self, kind='edge', prim=None, excludePatches=False):
		import networkx as nx

		#Have to use DiGraph in order to draw any arrows
		G = nx.DiGraph(name=kind)
		agents = list(self.allagents.values()) if prim is None else self.agents[prim]
		if excludePatches: agents = [a for a in agents if a.primitive!='patch']
		G.add_nodes_from([(a.id, {'breed': a.breed, 'primitive': a.primitive, 'position': None if a.position is None else a.position.copy()}) for a in agents])
		ae = self.allEdges
		if kind in ae:
			for e in ae[kind]:
				if prim is None or (e.vertices[0].primitive==prim and e.vertices[1].primitive==prim): G.add_edge(
					e.startpoint.id if e.directed else e.vertices[0].id,
					e.endpoint.id if e.directed else e.vertices[1].id,
					weight=e.weight, directed=e.directed
				)
		return G

	@property
	def allEdges(self):
		es = {}
		for a in self.allagents.values():
			for e in a.alledges:
				if not e.kind in es: es[e.kind] = []
				if not e in es[e.kind]: es[e.kind].append(e)
		return es

	def spatial(self, *args, **kwargs):
		from helipad.spatial import spatialSetup
		return spatialSetup(self, *args, **kwargs)

	@property
	def allagents(self):
		agents = {}
		for l in self.agents.values():
			agents.update({a.id:a for a in l})
		return agents

	#CALLBACK FOR DEFAULT PARAMETERS
	#Model param redundant, strictly speaking, but it's necessary to make the signature match the setter callback
	def nUpdater(self, val, prim, model, force=False):
		if not self.hasModel and not force: return val

		if 'num_' in prim: prim = prim.split('_')[1] #Because the parameter callback passes num_{prim}
		array = self.agents[prim]
		diff = val - len(array)

		#Add agents
		if diff > 0:
			ids = [a.id for a in self.allagents.values()]
			maxid = max(ids) if ids else 0 #Figure out maximum existing ID
			for aId in range(maxid+1, maxid+int(diff)+1):
				breed = self.doHooks([prim+'DecideBreed', 'decideBreed'], [aId, self.primitives[prim].breeds.keys(), self])
				if breed is None: breed = list(self.primitives[prim].breeds.keys())[aId%len(self.primitives[prim].breeds)]
				if not breed in self.primitives[prim].breeds:
					raise ValueError(ï('Breed \'{0}\' is not registered for the \'{1}\' primitive.').format(breed, prim))
				new = self.primitives[prim].class_(breed, aId, self)
				array.append(new)

		#Remove agents
		elif diff < 0:
			shuffle(array) #Delete agents at random

			#Remove agents, maintaining the proportion between breeds
			n = {x: 0 for x in self.primitives[prim].breeds.keys()}
			for a in self.agents[prim]:
				if n[a.breed] < -diff:
					n[a.breed] += 1
					a.die(updateGUI=False)
				else: continue

	#
	# DEBUG FUNCTIONS
	# Only call from the console, not in user code
	#

	# Requires to be run from Terminal (⌘-⇧-R in TextMate). `self` will refer to the model object
	# Readline doesn't look like it's doing anything here, but it enables certain console features
	# Only works on Mac. Also Gnureadline borks everything, so don't install that.
	# Has to be called *after* Cpanel.__init__() is called, or the cpanel object won't be available.
	def debugConsole(self):
		if sys.platform=='darwin' and isBuffered():
			try:
				import code, readline
				env = globals().copy()
				env['self'] = self
				code.interact(local=env)
			except ModuleNotFoundError: print(ï('Error initializing the debug console. Make sure the `readline` and `code` modules are installed.'))

	#Return agents of a breed if string; return specific agent with ID otherwise
	def agent(self, var, primitive=None):
		if primitive is None:
			primitive = 'agent' if 'agent' in self.primitives else next(iter(self.primitives))
		if isinstance(var, str):
			return [a for a in self.agents[primitive] if a.breed==var]
		else:
			aa = self.allagents
			return aa[var] if var in aa else None

		return None #If nobody matched

	#Returns summary statistics on an agent variable at a single point in time
	def summary(self, var, prim=None, breed=None, good=False):
		if prim is None:
			prim = 'agent' if 'agent' in self.primitives else next(iter(self.primitives))
		agents = self.agents[prim] if breed is None else self.agent(breed, prim)
		if not good: data = pandas.Series([getattr(a, var) for a in agents]) #Pandas gives us nice statistical functions
		else: data = pandas.Series([a.stocks[var] for a in agents])
		stats = {
			'n': data.size,
			'Mean': data.mean(),
			'StDev': data.std(),
			'Variance': data.var(),
			'Maximum': data.max(),
			'Minimum': data.min(),
			'Sum': data.sum()
		}
		for k, v in stats.items():
			print(k+': ', v)

	def launchCpanel(self):
		if hasattr(self, 'breed'): warnings.warn(ï('Control panel can only be launched on the top-level model.'), None, 2)

		self.doHooks('CpanelPreLaunch', [self])

		#Set our agents slider to be a multiple of how many agent types there are
		#Do this down here so we can have breeds registered before determining options
		for k,p in self.primitives.items():
			if self.params['num_'+k].type != 'hidden':
				l = len(p.breeds)
				if not l: continue
				self.params['num_'+k].opts['low'] = makeDivisible(self.params['num_'+k].opts['low'], l, 'max')
				self.params['num_'+k].opts['high'] = makeDivisible(self.params['num_'+k].opts['high'], l, 'max')
				self.params['num_'+k].opts['step'] = makeDivisible(self.params['num_'+k].opts['low'], l, 'max')
				self.params['num_'+k].value = makeDivisible(self.params['num_'+k].value, l, 'max')
				self.params['num_'+k].default = makeDivisible(self.params['num_'+k].default, l, 'max')

		if not isNotebook():
			from helipad.cpanelTkinter import Cpanel
			self.cpanel = Cpanel(self)
			self.doHooks('CpanelPostInit', [self.cpanel]) #Want the cpanel property to be available here, so don't put in cpanel.py
			self.debugConsole()
			self.cpanel.mainloop()		#Launch the control panel
		else:
			from helipad.cpanelJupyter import Cpanel, SilentExit
			if self.cpanel: self.cpanel.invalidate(ï('Control panel was redrawn in another cell.'))
			self.cpanel = Cpanel(self)
			self.doHooks('CpanelPostInit', [self.cpanel])
			raise SilentExit() #Don't blow past the cpanel if doing "run all"

		self.doHooks('GUIClose', [self]) #This only executes after all GUI elements have closed

	def launchVisual(self):
		if self.visual is None or self.visual.isNull:
			if not self.cpanel:
				print(ï('No visualizations available. To run the model with no GUI, use model.start() instead.'))
				return
			if not self.param('stopafter') or not (self.param('csv') or 'terminate' in self.hooks):
				print(ï('Running from the control panel with no visualization requires a stop condition, and either CSV export or a terminate hook.'))
				return

		self.setup()

		if self.visual is not None and not self.visual.isNull:
			self.visual.launch(ï('{}Data Plots').format(self.name+(' ' if self.name!='' else '')))
		else: #Headless
			self.params['stopafter'].disable()
			self.params['csv'].disable()

		self.doHooks('visualLaunch', [self, self.visual])

		#If we're running in cpanel-less mode, hook through a Tcl loop so it doesn't exit on pause
		if not self.cpanel and not self.visual.isNull and not isNotebook():
			from tkinter import Tcl
			root = Tcl()
			root.after(1, self.start)
			self.debugConsole()
			root.mainloop()
		else: self.start() #As long as we haven't already started

	# Generates function decorators for hooks, reporters, etc.
	def genDecorator(self, todo):
		def dec(name=None, **kwargs):
			if callable(name): func, name, isDec = (name, None, True)
			else: isDec = False

			#Is a decorator
			def rep1(fn):
				namn = name #If I do anything with name here, Python throws UnboundLocalError. Possibly a Python bug?
				if namn is None: namn=fn.__name__
				todo(namn, fn, kwargs)
				return fn

			return rep1(func) if isDec else rep1
		return dec

	#===================
	# DEPRECATED METHODS
	#===================

	# DEPRECATED IN HELIPAD 1.4; REMOVE IN HELIPAD 1.6

	def addEvent(self, name, fn, **kwargs):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.addEvent()', 'model.events.add()'), FutureWarning, 2)
		return self.events.add(name, fn, **kwargs)

	def removeEvent(self, name):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.removeEvent()', 'model.events.remove()'), FutureWarning, 2)
		return self.events.remove(name)

	def clearEvents(self):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.clearEvents()', 'model.events.clear()'), FutureWarning, 2)
		self.events.clear()

	def addParameter(self, *args, **kwargs):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.addParameter()', 'model.params.add()'), FutureWarning, 2)
		return self.params.add(*args, **kwargs)

	def addBreedParam(self, name, title, type, dflt, opts={}, prim=None, runtime=True, callback=None, desc=None, getter=None, setter=None):
		if prim is None:
			if len(self.primitives) == 1: prim = next(iter(self.primitives.keys()))
			else: raise KeyError(ï('Breed parameter must specify which primitive it belongs to.'))
		return self.addParameter(name, title, type, dflt, opts, runtime, callback, 'breed', desc, prim=prim, getter=getter, setter=setter)

	def addGoodParam(self, name, title, type, dflt, opts={}, runtime=True, callback=None, desc=None, getter=None, setter=None):
		return self.addParameter(name, title, type, dflt, opts, runtime, callback, 'good', desc, getter=getter, setter=setter)

	@property
	def goodParams(self):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.goodParams', 'model.params.perGood'), FutureWarning, 2)
		return self.params.perGood

	@property
	def allParams(self):
		warnings.warn(ï('Model.allParams is deprecated. All parameters can be accessed using model.params.'), FutureWarning, 2)
		return self.params.values()

	def addHook(self, place, func, prioritize=False):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.addHook()', 'model.hooks.add()'), FutureWarning, 2)
		return self.hooks.add(place, func, prioritize)

	def removeHook(self, place, fname, removeall=False):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.removeHook()', 'model.hooks.remove()'), FutureWarning, 2)
		return self.hooks.remove(place, fname, removeall=False)

	def clearHooks(self, place):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.clearHooks()', 'model.hooks.clear()'), FutureWarning, 2)
		return self.hooks.remove(place)

	def addPrimitive(self, *args, **kwargs):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.addPrimitive()', 'model.primitives.add()'), FutureWarning, 2)
		return self.primitives.add(*args, **kwargs)

	def removePrimitive(self, name):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.removePrimitive()', 'model.primitives.remove()'), FutureWarning, 2)
		return self.primitives.remove(name)

	def addGood(self, *args, **kwargs):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.addGood()', 'model.goods.add()'), FutureWarning, 2)
		return self.goods.add(*args, **kwargs)

	@property
	def moneyGood(self):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.moneyGood', 'model.goods.money'), FutureWarning, 2)
		return self.goods.money

	@property
	def nonMoneyGoods(self):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('Model.nonMoneyGoods()', 'model.goods.nonmonetary'), FutureWarning, 2)
		return self.goods.nonmonetary

class MultiLevel(baseAgent, Helipad):
	def __init__(self, breed, id, parentModel):
		super().__init__(breed, id, parentModel)
		self.setup()

	#Deprecated in Helipad 1.4; remove in Helipad 1.6
	@property
	def dontStepAgents(self):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('MultiLevel.dontStepAgents', 'MultiLevel.cutStep()'), FutureWarning, 2)
		return self._cut

	@dontStepAgents.setter
	def dontStepAgents(self, val):
		warnings.warn(ï('{0} is deprecated and has been replaced with {1}.').format('MultiLevel.dontStepAgents', 'MultiLevel.cutStep()'), FutureWarning, 2)
		self._cut = val

#==================
# CONTAINER CLASSES
#==================

#For adding breeds and goods. Should not be called directly
class gandb(funcStore):
	def __init__(self, model):
		self.model = model

	def add(self, obj, name, color, prim=None, **kwargs):
		if name in self:
			warnings.warn(ï('{0} \'{1}\' already defined. Overriding…').format(obj, name), None, 2)

		cobj = color if isinstance(color, Color) else Color(color)
		cobj2 = cobj.lighten()
		self[name] = Item(color=cobj, color2=cobj2, **kwargs)

		#Make sure the parameter lists keep up with our items
		if obj=='good': paramDict = self.model.params.perGood
		elif obj=='breed': paramDict = {k:v for k,v in self.model.params.perBreed.items() if v.prim==self.primitive}
		for p in paramDict.values(): p.addKey(name)

		return self[name]

	def remove(self, name):
		if isinstance(name, (list, tuple)): return [self.remove(n) for n in name]
		if not name in self: return False

		#Also delete per-item parameters
		pdict = self.model.params.perBreed if isinstance(self, Breeds) else self.model.params.perGood
		for param in pdict.values():
			del param.value[name]
			if getattr(param, 'elements', False):
				if not isNotebook(): param.elements[name].destroy()
				else: param.elements[name].close()
				del param.elements[name]
		return super().remove(name)

class Goods(gandb):
	def add(self, name, color, endowment=None, money=False, props={}):
		if money:
			if self.money is not None:
				print(ï('Money good already specified as {}. Overriding…').format(self.money))
				self[self.money].money = False

			#Add the M0 plot once we have a money good, only if we haven't done it before
			elif (self.model.visual is None or self.model.visual.isNull) and hasattr(self.model.visual, 'plots'):
				try:
					if not 'money' in self.model.visual.plots: self.model.visual.addPlot('money', ï('Money'), selected=False)
				except: pass #Can't add plot if re-drawing the cpanel

		props['quantity'] = endowment
		item = super().add('good', name, color, money=money, props=props)

		#Add demand plot once we have at least 2 goods
		if len(self) == 2 and (self.model.visual is None or self.model.visual.isNull) and hasattr(self.model.visual, 'plots'):
			try:
				if not 'demand' in self.model.visual.plots: self.model.visual.addPlot('demand', ï('Demand'), selected=False)
			except: pass

		return item

	@property
	def money(self):
		for name,good in self.items():
			if good.money: return name
		return None

	@property
	def nonmonetary(self):
		return {k:v for k,v in self.items() if not v.money}

class Breeds(gandb):
	def __init__(self, model, primitive):
		self.primitive = primitive
		super().__init__(model)

	def add(self, name, color): return super().add('breed', name, color)

class Primitives(funcStore):
	def __init__(self, model):
		self.model = model
		super().__init__()

	def add(self, name, class_, plural=None, dflt=50, low=1, high=100, step=1, hidden=False, priority=100, order=None):
		if name=='all': raise ValueError(ï('{} is a reserved name. Please choose another.').format(name))
		if not plural: plural = name+'s'
		class_.primitive = name
		self[name] = Item(
			class_=class_,
			plural=plural,
			priority=priority,
			order=order,
			breeds=Breeds(self.model, name)
		)
		def popget(name, model):
			prim = name.split('_')[1]
			if not model.hasModel: return None
			else: return len(model.agents[prim])

		self.model.params.add('num_'+name, 'Number of '+plural.title(), 'hidden' if hidden else 'slider', dflt=dflt, opts={'low': low, 'high': high, 'step': step} if not hidden else None, setter=self.model.nUpdater, getter=popget)
		self.model.agents[name] = []

	def remove(self, name):
		val = super().remove(name)
		if val:
			del self.model.agents[name]
			del self.model.params['num_'+name]
		return val

class Events(funcStore):
	def __init__(self):
		super().__init__()
		class Event:
			def __init__(self, name, trigger, repeat=False, **kwargs):
				self.name = name
				self.trigger = trigger
				self.repeat = repeat
				self.kwargs = kwargs
				self.data = []
				self.triggered = []
				self.reset()

			def check(self, model):
				if self.triggered and not self.repeat: return False
				if (isinstance(self.trigger, int) and model.t==self.trigger) or (callable(self.trigger) and self.trigger(model)):
					data = {k: v[0] for k,v in model.data.getLast(1).items()}
					if self.repeat:
						self.data.append(data)
						self.triggered.append(model.t)
					else:
						self.data = data
						self.triggered = model.t
					return True

			def reset(self):
				if self.repeat:
					self.data.clear()
					self.triggered.clear()
				else:
					self.data = None
					self.triggered = False
		self.Event = Event

	def add(self, name, function, **kwargs):
		return super().add(name, self.Event(name, function, **kwargs))

class Hooks(funcStore):
	multi = True

	def add(self, name, function, prioritize=False):
		if not name in self: self[name] = []
		if prioritize: self[name].insert(0, function)
		else: self[name].append(function)


#===============
# Functions to initialize a grid network of Patch agents
# and add orientation and motion functions to other agents.
#===============

import warnings
from random import uniform
from math import sqrt, sin, cos, atan2, pi, copysign, floor
from helipad.agent import Patch, baseAgent
from helipad.visualize import Charts
from helipad.helpers import ï, Number

#===============
# SETUP
# Create parameters, add functions, and so on
#===============

def spatialSetup(model, dim=10, corners=False, geometry='rect', **kwargs):
	# Backward compatibility
	if 'x' in kwargs:		#Remove in Helipad 1.6
		dim = kwargs['x'] if 'y' not in kwargs else (kwargs['x'], kwargs['y'])
		warnings.warn(ï('Using x and y to set dimensions is deprecated. Use the dim argument instead.'), FutureWarning, 3)
	if 'diag' in kwargs:	#Remove in Helipad 1.7
		corners = kwargs['diag']
		warnings.warn(ï('The `diag` argument is deprecated. Use the `corners` argument instead.'), FutureWarning, 3)

	#Initialize patch container and primitive
	model.primitives.add('patch', Patch, hidden=True, priority=-10)
	pClasses = {'rect': PatchesRect, 'polar': PatchesPolar}
	if not isinstance(geometry, str): #We're gonna throw an error anyway if it's not a class or a string
		pClasses[geometry.geometry] = geometry
		geometry = geometry.geometry
	model.patches = pClasses[geometry](dim, **kwargs)
	def npsetter(val, item): raise RuntimeError(ï('Patch number cannot be set directly. Set the dim parameter instead.'))
	model.params['num_patch'].getter = lambda item: len(model.patches)
	model.params['num_patch'].setter = npsetter

	#Dimension parameters. Deprecated in Helipad 1.5; remove in Helipad 1.7
	def dimgen(dim, action):
		def dimget(name, model):
			warnings.warn(ï("The {0} parameter is deprecated. Use {1} instead.").format(f"'{dim}'", "model.patches.dim"), FutureWarning, 5)
			return model.patches.dim[0 if dim=='x' else 1]
		def dimset(val, name, model):
			warnings.warn(ï("The {0} parameter is deprecated. Use {1} instead.").format(f"'{dim}'", "model.patches.dim"), FutureWarning, 5)
			model.patches.dim[0 if dim=='x' else 1] = val
		return dimget if action=='get' else dimset
	def wrapget(name, model):
		warnings.warn(ï("The {0} parameter is deprecated. Use {1} instead.").format("'wrap'", "model.patches.wrap"), FutureWarning, 5)
		return model.patches.wrap == (True, True)
	def wrapset(val, name, model):
		warnings.warn(ï("The {0} parameter is deprecated. Use {1} instead.").format("'wrap'", "model.patches.wrap"), FutureWarning, 5)
		model.patches.wrap = val if isinstance(val, (tuple, list)) else (val, val)
	wrap = kwargs['wrap'] if 'wrap' in kwargs else True
	model.params.add('x', ï('Map Width'), 'hidden', dflt=model.patches.dim[0], setter=dimgen('x','set'), getter=dimgen('x','get'))
	model.params.add('y', ï('Map Height'), 'hidden', dflt=model.patches.dim[1], setter=dimgen('y','set'), getter=dimgen('y','get'))
	model.params.add('wrap', ï('Wrap'), 'hidden', dflt=(wrap is True), setter=wrapset, getter=wrapget) #Only checked at the beginning of a model

	#Hook a positioning function or randomly position our agents
	@model.hook(prioritize=True)
	def baseAgentInit(agent, model):
		if agent.primitive == 'patch': return #Patch position is fixed
		p = model.doHooks(['baseAgentPosition', agent.primitive+'Position'], [agent, agent.model])
		agent.position = list(p) if p and len(p) >= 2 else [uniform(*model.patches.boundaries[0]), uniform(*model.patches.boundaries[1])]

	#MOVEMENT AND POSITION

	#Both agents and patches have x and y properties
	def setx(self, val): self.moveTo(val, self.position[1])
	def sety(self, val): self.moveTo(self.position[0], val)
	baseAgent.x = property(lambda self: self.position[0], setx)
	baseAgent.y = property(lambda self: self.position[1], sety)

	def move(self, x, y):
		mapx, mapy = self.model.patches.dim
		xlim, ylim = self.model.patches.boundaries
		self.position[0] += x
		self.position[1] += y

		if self.model.patches.wrap[0]:
			while self.position[0] >= xlim[1]: self.position[0] -= mapx
			while self.position[0] < xlim[0]: self.position[0] += mapx
		else:
			self.position[0] = min(self.position[0], xlim[1])
			self.position[0] = max(self.position[0], xlim[0])

		if self.model.patches.wrap[1]:
			while self.position[1] >= ylim[1]: self.position[1] -= mapy
			while self.position[1] < ylim[0]: self.position[1] += mapy
		else:
			self.position[1] = min(self.position[1], ylim[1])
			self.position[1] = max(self.position[1], ylim[0])

	baseAgent.move = NotPatches(move)

	#Can also take an agent or patch as a single argument
	def moveTo(self, x, y=None):
		if not isinstance(x, Number): x,y = x.position
		if x>self.model.patches.boundaries[0][1] or y>self.model.patches.boundaries[1][1] or x<self.model.patches.boundaries[0][0] or y<self.model.patches.boundaries[1][0]:
			raise IndexError(ï('Dimension is out of range.'))
		self.position = [x,y]
	baseAgent.moveTo = NotPatches(moveTo)

	#Initialize the patch container at the beginning of the model
	@model.hook(prioritize=True)
	def modelPreSetup(model):
		model.patches.clear()

	#Position our patches in the coordinate system
	@model.hook(prioritize=True)
	def patchInit(agent, model):
		model.patches.place(agent)
		agent.colorData = {}

	#Establish grid links
	@model.hook(prioritize=True)
	def modelPostSetup(model):
		for patch in model.agents['patch']:
			neighbors = model.patches.neighbors(patch, corners)
			connections = patch.neighbors #Neighbors that already have a connection
			for n, weight in neighbors:
				if not n in connections:
					patch.newEdge(n, 'space', weight=weight)

	#Don't reset the visualizer if charts is already registered
	if model.visual is None or not isinstance(model.visual, Charts):
		model.useVisual(Charts)

	mapPlot = model.visual.addPlot('map', 'Map', type='network', layout='patchgrid', projection=geometry if geometry=='polar' else None)
	return mapPlot

#Functions for all primitives except patches, which are spatially fixed
def NotPatches(function):
	def np2(self, *args, **kwargs):
		if self.primitive != 'patch': return function(self, *args, **kwargs)
		else: raise RuntimeError(ï('Patches cannot move.'))
	return np2

#===============
# COORDINATE SYSTEMS
#===============

#A 2D list of lists that lets us use [x,y] indices
class Patches2D(list):
	def __init__(self, dim, props=[], funcs=[], **kwargs):
		if isinstance(dim, int): dim = (dim, dim)
		if len(dim) != 2: raise TypeError(ï('Invalid dimension.'))
		self.dim = dim
		super().__init__([])

		#Attach coordinate-specific functions and properties to agent objects
		for p in props: setattr(Patch, p.__name__, property(p))
		for f in funcs: setattr(baseAgent, f.__name__, NotPatches(f))

	def __getitem__(self, key):
		if isinstance(key, int): return super().__getitem__(key)
		else:
			if key[1] is None: return super().__getitem__(key[0])
			if key[0] is None: return [x[key[1]] for x in self]
			return super().__getitem__(key[0])[key[1]]

	def __setitem__(self, key, val):
		if isinstance(key, int): return super().__setitem__(key, val)
		else: super().__getitem__(key[0])[key[1]] = val

	def __repr__(self): return f'<{self.__class__.__name__}: {self.dim[0]}×{self.dim[1]}>'

#Each row and column of equal length
class PatchesRect(Patches2D):
	geometry = 'rect'

	#Give patches and agents methods for their neighbors in the four cardinal directions
	def __init__(self, dim, wrap=True, **kwargs):
		def up(patch):
			if patch.y==0 and not self.wrap[1]: return None
			return self[patch.x, patch.y-1 if patch.y > 0 else self.dim[1]-1]

		def right(patch):
			if patch.x>=self.dim[0]-1 and not self.wrap[0]: return None
			return self[patch.x+1 if patch.x < self.dim[0]-1 else 0, patch.y]

		def down(patch):
			if patch.y>=self.dim[1]-1 and not self.wrap[1]: return None
			return self[patch.x, patch.y+1 if patch.y < self.dim[1]-1 else 0]

		def left(patch):
			if patch.x==0 and not self.wrap[0]: return None
			return self[patch.x-1 if patch.x > 0 else self.dim[0]-1, patch.y]

		def agentsOn(patch):
			for prim, lst in patch.model.agents.items():
				if prim=='patch': continue
				yield from [a for a in lst if patch.x-0.5<=a.x<patch.x+0.5 and patch.y-0.5<=a.y<patch.y+0.5]

		baseAgent.patch = property(lambda agent: self[round(agent.x), round(agent.y)])

		def moveUp(agent): agent.move(0, -1)
		def moveRight(agent): agent.move(1, 0)
		def moveDown(agent): agent.move(0, 1)
		def moveLeft(agent): agent.move(-1, 0)
		def forward(self, steps=1):
			self.move(steps * cos(self.rads-pi/2), steps * sin(self.rads-pi/2))

		#Internal function  to calculate x,y distances s.t. wrap constraints for distance and orientation purposes
		def _offset(agent, x, y):
			difx, dify = x - agent.x, y - agent.y
			if self.wrap[0]:
				crossx = difx - copysign(self.dim[0], difx)
				if abs(crossx) < abs(difx): difx = crossx
			if self.wrap[1]:
				crossy = dify - copysign(self.dim[1], dify)
				if abs(crossy) < abs(dify): dify = crossy
			return (difx, dify)
		baseAgent._offset = _offset

		def distanceFrom(agent, agent2):
			difx, dify = agent._offset(agent2.x, agent2.y)
			return sqrt(difx**2 + dify**2)
		baseAgent.distanceFrom = distanceFrom

		def orientTo(agent, x, y=None):
			if not isinstance(x, Number): x,y = x.position
			difx, dify = agent._offset(x, y)
			agent.rads = atan2(dify,difx) + 0.5*pi #atan calculates 0º pointing East. We want 0º pointing North

		if isinstance(wrap, bool): wrap = (wrap, wrap)
		if len(wrap) != 2: raise TypeError(ï('Invalid wrap parameter.'))
		self.wrap = wrap
		super().__init__(dim, wrap=wrap, props=[up, right, down, left, agentsOn], funcs=[moveUp, moveRight, moveDown, moveLeft, forward, orientTo], **kwargs)

	def neighbors(self, patch, corners):
		neighbors = [(patch.up, 1), (patch.right, 1), (patch.down, 1), (patch.left, 1)]
		if corners: neighbors += [(patch.up.left, corners), (patch.right.up, corners), (patch.down.right, corners), (patch.left.down, corners)]
		return [p for p in neighbors if p[0] is not None and p[0] is not self]

	#Take a sequential list of self.count patches and position them appropriately in the internal list
	def place(self, agent):
		if not super().__len__(): self += [[] for i in range(self.dim[0])]
		x=0
		while len(self[x]) >= self.dim[1]: x+=1		#Find a column that's not full yet
		agent.position = (x, len(self[x]))			#Note the position
		self[x].append(agent)						#Append the agent

	@property #((xmin, xmax),(ymin, ymax))
	def boundaries(self): return ((-0.5, self.dim[0]-0.5), (-0.5, self.dim[1]-0.5))

	def __len__(self): return self.dim[0] * self.dim[1]

#x,y placement is the same; just need to redefine patch functions and neighbors
#x is number around, y is number out
class PatchesPolar(PatchesRect):
	geometry = 'polar'
	wrap = (True, False) #Wrap around, but not in-to-out.

	def __init__(self, dim, **kwargs):
		def inward(patch):
			if patch.y==0: return None
			return self[patch.x, patch.y-1]

		def outward(patch):
			if patch.y>=self.dim[1]-1: return None
			return self[patch.x, patch.y+1]

		def clockwise(patch):
			return self[patch.x+1 if patch.x < self.dim[0]-1 else 0, patch.y]

		def counterclockwise(patch):
			return self[patch.x-1 if patch.x > 0 else self.dim[0]-1, patch.y]

		def agentsOn(patch):
			for prim, lst in patch.model.agents.items():
				if prim=='patch': continue
				yield from [a for a in lst if floor(a.x)==patch.x and floor(a.y)==patch.y]

		baseAgent.patch = property(lambda agent: self[floor(agent.x), floor(agent.y)])

		def moveInward(agent): agent.move(0, -1)
		def moveOutward(agent): agent.move(0, 1)
		def moveClockwise(agent): agent.move(1, 0)
		def moveCounterclockwise(agent): agent.move(-1, 0)
		def forward(agent, steps=1):
			#2π-self.rads to make it go clockwise
			th1 = 2*pi-(2*pi/self.dim[0] * agent.x)
			newth = th1 + atan2(steps*sin(2*pi-agent.rads-th1), agent.y+steps*cos(2*pi-agent.rads-th1))
			newx = (2*pi-newth)/(2*pi/self.dim[0])
			if newx > self.dim[0]: newx -= self.dim[0]
			newr = sqrt(agent.y**2 + steps**2 + 2*agent.y*steps*cos(2*pi-agent.rads-th1))
			agent.move(newx-agent.x, newr-agent.y)

		def distanceFrom(agent, agent2):
			th1 = agent.x * 2*pi /self.dim[0]
			th2 = agent2.x * 2*pi /self.dim[0]
			return sqrt(agent.y**2 + agent2.y**2 - 2*agent.y*agent2.y*cos(th1-th2))
		baseAgent.distanceFrom = distanceFrom

		def orientTo(agent, x, y=None):
			if not isinstance(x, Number): x,y = x.position
			th1 = 2*pi-(2*pi/self.dim[0] * agent.x)
			th2 = 2*pi-(2*pi/self.dim[0] * x)
			agent.rads = pi - atan2(agent.y * sin(th1) - y * sin(th2), agent.y * cos(th1) - y * cos(th2))

		#Skip PatchesRect.__init__
		Patches2D.__init__(self, dim, props=[inward, outward, clockwise, counterclockwise, agentsOn], funcs=[moveInward, moveOutward, moveClockwise, moveCounterclockwise, forward, orientTo])

	#The usual 3-4 neighbors, but if corners are on, all patches in the center ring will be neighbors
	def neighbors(self, patch, corners):
		neighbors = [(patch.inward, 1), (patch.outward, 1), (patch.clockwise, 1), (patch.counterclockwise, 1)]
		if corners:
			neighbors += [(patch.clockwise.inward, corners), (patch.counterclockwise.inward, corners), (patch.clockwise.outward, corners), (patch.counterclockwise.outward, corners)]
			if patch.y==0:
				flatneighbors = [n[0] for n in neighbors]
				for p in patch.model.patches[None, 0]:
					if p not in flatneighbors:
						neighbors.append((p, corners))

		return [p for p in neighbors if p[0] is not None and p[0] is not self]

	@property #((xmin, xmax),(ymin, ymax))
	def boundaries(self): return ((0, self.dim[0]), (0, self.dim[1]))


# ==========
# Basic extensible agent class
# Do not run this file; import model.py and run from your file.
# ==========

import warnings
from random import choice, randint
from math import degrees, radians, pi
import numpy as np
from helipad.helpers import ï, Number

#Basic agent functions. This class should not be instantiated directly; instead it should be
#subclassed by a class corresponding to a primitive and registered with Helipad.addPrimitive().
#See below, the Agent() class for a minimal example.
class baseAgent:
	angleUnit = 'deg'
	fixed = False
	overdraft = 'continue-silent'

	#==================
	# BASIC METHODS
	#==================

	def __init__(self, breed, aId, model):
		self.breed = breed
		self.id = int(aId)
		self.model = model
		self.age = 0
		self.dead = False
		self.stocks = Stocks(breed, model.goods)
		self.edges = {}
		self.utils = 0
		self.position = None #Overridden in spatial init
		self.currentDemand = {g:0 for g in model.goods.keys()}
		self.rads = 0

		#For multi-level models
		#Has to be a static property since we're checking as the object initializes
		if hasattr(super(), 'runInit'): super().__init__()

		self.model.doHooks(['baseAgentInit', self.primitive+'Init'], [self, self.model])

	def __repr__(self): return f'<{self.__class__.__name__} {self.id}>'

	def step(self, stage:Number):
		self.model.doHooks(['baseAgentStep', self.primitive+'Step'], [self, self.model, stage])
		if hasattr(super(), 'runInit'): super().step(stage) #For multi-level models
		if stage == self.model.stages: self.age += 1

	#==================
	# ECONOMIC METHODS
	#==================

	#Give amt1 of good 1, get amt2 of good 2
	#Negative values of amt1 and amt2 allowed, which reverses the direction
	def trade(self, partner:baseAgent, good1:str, amt1:Number, good2:str, amt2:Number):
		go = True
		self.model.doHooks('preTrade', [self, partner, good1, amt1, good2, amt2])

		#Budget constraints. Hold price constant if hit
		message = None
		prim1, prim2 = self.primitive.title(), partner.primitive.title()
		if amt2 != 0: price = amt1 / amt2
		if self.overdraft != 'allow':
			if amt1 > self.stocks[good1]:
				message = ï('{0} {1} does not have sufficient {2} to give {3} {4}.').format(prim1, self.id, good1, prim2, partner.id)
				if 'continue' in self.overdraft:
					message += ï(' Continuing with available {0} of {1}…').format(good1, self.stocks[good1])
					amt1 = self.stocks[good1]
					if amt2 != 0: amt2 = amt1 / price
			elif -amt1 > partner.stocks[good1]:
				message = ï('{0} {1} does not have sufficient {2} to give {3} {4}.').format(prim2, partner.id, good1, prim1, self.id)
				if 'continue' in self.overdraft:
					message += ï(' Continuing with available {0} of {1}…').format(good1, partner.stocks[good1])
					amt1 = -partner.stocks[good1]
					if amt2 != 0: amt2 = amt1 / price
			if amt2 > partner.stocks[good2]:
				message = ï('{0} {1} does not have sufficient {2} to give {3} {4}.').format(prim2, partner.id, good2, prim1, self.id)
				if 'continue' in self.overdraft:
					message += ï(' Continuing with available {0} of {1}…').format(good2, partner.stocks[good2])
					amt2 = partner.stocks[good2]
					amt1 = price * amt2
			elif -amt2 > self.stocks[good2]:
				message = ï('{0} {1} does not have sufficient {2} to give {3} {4}.').format(prim1, self.id, good2, prim2, partner.id)
				if 'continue' in self.overdraft:
					message += ï(' Continuing with available {0} of {1}…').format(good2, self.stocks[good2])
					amt2 = -self.stocks[good2]
					amt1 = price * amt2

			if message is not None:
				if self.overdraft == 'stop': raise ValueError(message)
				if 'fail' in self.overdraft:
					go = False
					message += ï(' Cancelling trade…')
				elif 'warn' in self.overdraft: warnings.warn(message, None, 2)

		if go:
			self.stocks[good1] -= amt1
			partner.stocks[good1] += amt1
			self.stocks[good2] += amt2
			partner.stocks[good2] -= amt2

			#Record demand
			if amt1 > 0: partner.currentDemand[good1] += amt1
			else: self.currentDemand[good1] -= amt1
			if amt2 > 0: self.currentDemand[good2] += amt2
			else: partner.currentDemand[good2] -= amt2

		self.model.doHooks('postTrade', [self, partner, good1, amt1, good2, amt2])

	#Price is per-unit
	#Returns the quantity actually sold, Which is the same as quantity input unless there's a shortage
	def buy(self, partner:baseAgent, good:str, q:Number, p:Number) -> Number:
		if self.model.goods.money is None: raise RuntimeError(ï('{} requires a monetary good to be specified.').format('Agent.buy()'))
		qp = self.model.doHooks('buy', [self, partner, good, q, p])
		if qp is not None: q, p = qp

		before = self.stocks[good]
		self.trade(partner, self.model.goods.money, p*q, good, q)
		return self.stocks[good] - before

	#Unilateral
	def pay(self, recipient:baseAgent, amount:Number) -> Number:
		if self.model.goods.money is None: raise RuntimeError(ï('{} requires a monetary good to be specified.').format('Agent.pay()'))
		go = True

		#Do hooks before budget constraints
		amount_ = self.model.doHooks('pay', [self, recipient, amount, self.model])
		if amount_ is not None: amount = amount_

		#Budget constraints
		prim1, prim2 = self.primitive.title(), recipient.primitive.title()
		message = None
		if self.overdraft != 'allow':
			if amount > self.balance:
				message = ï('{0} {1} does not have sufficient funds to pay {2} {3}.').format(prim1, self.id, prim2, recipient.id)
				if self.overdraft == 'stop': raise ValueError(message)
				if 'continue' in self.overdraft:
					amount = self.balance
					message += ï(' Continuing with available balance of {}…').format(self.balance)
			elif -amount > recipient.balance:
				message = ï('{0} {1} does not have sufficient funds to pay {2} {3}.').format(prim2, recipient.id, prim1, self.id)
				if 'continue' in self.overdraft:
					amount = -recipient.balance
					message += ï(' Continuing with available balance of {}…').format(recipient.balance)

			if message is not None:
				if self.overdraft == 'stop': raise ValueError(message)
				if 'fail' in self.overdraft:
					go = False
					message += ï(' Cancelling trade…')
				if 'warn' in self.overdraft:
					warnings.warn(message, None, 2)

		if go and amount:
			recipient.stocks[self.model.goods.money] += amount
			self.stocks[self.model.goods.money] -= amount
			return amount
		else: return 0

	@property
	def balance(self):
		if self.model.goods.money is None: raise RuntimeError(ï('Balance checking requires a monetary good to be specified.'))
		bal = self.stocks[self.model.goods.money]
		bal_ = self.model.doHooks('checkBalance', [self, bal, self.model])
		if bal_ is not None: bal = bal_

		return bal

	#==================
	# GENETIC METHODS
	#==================

	def reproduce(self, inherit=[], mutate={}, partners=[]):
		if self.fixed: raise NotImplementedError(ï('Fixed primitives cannot reproduce.'))

		maxid = 0
		for a in self.model.allagents.values():
			if a.id > maxid:
				maxid = a.id
		newagent = type(self)(self.breed, maxid+1, self.model)

		#Values in the inherit list can either be a variable name (in which case the new agent inherits
		#the mean of all of the values for the parents), or a tuple, the first element of which is a
		#variable name, and the second is a string representing how to merge them. Possible values are
		#'mean' (default for numeric values), 'first' (default for non-numeric values), 'last', 'gmean',
		#'random', and 'sum'. The second value can also take a function, which receives a list of
		#values from the parents and returns a value for the child.
		parents = [self] + partners
		for a in inherit:
			stat = None
			if isinstance(a, tuple): a, stat = a
			v = [getattr(p,a) for p in parents if hasattr(p,a)] #List of values, filtering those without
			if len(v)==0: continue

			#Default statistic if unspecified. 'mean' for numbers, and 'first' for non-numbers.
			if stat is None:
				stat = 'mean' if isinstance(v[0], (int, float, complex)) and not isinstance(v[0], bool) else 'first'

			if stat=='mean': n = np.mean(v)
			elif stat=='sum': n = sum(v)
			elif stat=='gmean': n = np.exp(np.log(v).sum()/len(v))
			elif stat=='first': n = v[0]
			elif stat=='last': n = v[len(v)-1]
			elif stat in ('rand', 'random'): n = choice(v)
			elif stat=='max': n = max(v)
			elif stat=='min': n = min(v)
			elif callable(stat): n = stat(v)
			else: raise ValueError(ï('Invalid statistic {}.').format(stat))

			setattr(newagent, a, n)

		#Mutate variables
		#Values in the mutate dict can be either a function (which takes a value and returns a value),
		#  a number (a std dev by which to mutate the value), or a tuple, the first element of which
		#  is a std dev and the second of which is either 'log' or 'linear'
		for k,v in mutate.items():
			if callable(v): newval = v(getattr(newagent, k))
			else:
				if isinstance(v, tuple): v, scale = v
				else: scale = 'linear'

				if scale=='log': newval = np.random.lognormal(np.log(getattr(newagent, k)), v)
				else: newval = np.random.normal(getattr(newagent, k), v)
			setattr(newagent, k, newval)

		newagent.id = maxid+1
		for p in parents:
			p.newEdge(newagent,'lineage', True) #Keep track of parent-child relationships
		self.model.agents[self.primitive].append(newagent)

		self.model.doHooks(['baseAgentReproduce', self.primitive+'Reproduce'], [parents, newagent, self.model])
		return newagent

	def die(self, updateGUI=True):
		if self.fixed: raise NotImplementedError(ï('Fixed primitives cannot die.'))
		self.model.agents[self.primitive].remove(self)
		for edge in self.alledges: edge.cut()
		self.dead = True
		self.model.doHooks(['baseAgentDie', self.primitive+'Die'], [self])

	@property
	def parent(self):
		p = self.inbound('lineage', obj='agent')
		if len(p)==0: return None
		elif len(p)==1: return p[0]
		else: return p

	@property
	def children(self):
		return [edge.partner(self) for edge in self.outbound('lineage')]

	#==================
	# NETWORK METHODS
	#==================

	def newEdge(self, partner, kind='edge', direction=None, weight=1):
		return Edge(self, partner, kind, direction, weight)

	def outbound(self, kind='edge', undirected=False, obj='edge'):
		if obj not in ['agent', 'edge']: raise ValueError(ï('Object must be specified either \'agent\' or \'edge\'.'))
		if kind is None: edges = self.alledges
		else:
			if kind not in self.edges: return []
			edges = self.edges[kind]
		ob = [edge for edge in edges if edge.startpoint == self or (undirected and not edge.directed)]
		return ob if obj=='edge' else [e.partner(self) for e in ob]

	def inbound(self, kind='edge', undirected=False, obj='edge'):
		if obj not in ['agent', 'edge']: raise ValueError(ï('Object must be specified either \'agent\' or \'edge\'.'))
		if kind is None: edges = self.alledges
		else:
			if kind not in self.edges: return []
			edges = self.edges[kind]
		ib = [edge for edge in edges if edge.endpoint == self or (undirected and not edge.directed)]
		return ib if obj=='edge' else [e.partner(self) for e in ib]

	def edgesWith(self, partner, kind='edge'):
		if kind is not None:
			if kind not in self.edges: return []
			edges = self.edges[kind]
		else: edges = self.alledges
		return [edge for edge in edges if self in edge.vertices and partner in edge.vertices]

	@property
	def alledges(self):
		edges = []
		for e in self.edges.values(): edges += e
		return edges

	#==================
	# OTHER METHODS
	#==================

	#Agent.orientation reports and sets degrees or radians, depending on Agent.angleUnit.
	#Agent.rads always reports radians and is not documented.
	@property
	def orientation(self):
		if self.primitive == 'patch': return None
		return degrees(self.rads) if 'deg' in self.angleUnit else self.rads

	@orientation.setter
	def orientation(self, val):
		if self.primitive == 'patch': raise RuntimeError(ï('Patches cannot rotate.'))
		if 'deg' in self.angleUnit: val = radians(val)
		while val >= 2*pi: val -= 2*pi
		while val < 0: val += 2*pi
		self.rads = val

	def rotate(self, angle): self.orientation += angle

	#In a multi-level model, allow the agent to move to a different deme/firm/etc
	def transfer(self, dest):
		origin = self.model
		dest.agents[self.primitive].append(self)
		self.model = dest
		origin.agents[self.primitive].remove(self)
		self.model.doHooks(['baseAgentMove', self.primitive+'Move'], [self, origin, dest])

#The default agent class corresponding to the 'agent' primitive.
class Agent(baseAgent):
	pass

#For spatial models
class Patch(baseAgent):
	fixed = True

	@property
	def neighbors(self):
		return self.outbound('space', True, obj='agent')

	def __repr__(self): return f'<Patch at {self.position[0]},{self.position[1]}>'

#Direction can take an Agent object (corresponding to the endpoint),
#an int (0 for undirected, >0 for agent1 to agent2, and <0 for agent2 to agent1),
#or a boolean (False for undirected, True for agent1 to agent2)
class Edge:
	def __init__(self, agent1, agent2, kind='edge', direction=None, weight=1):
		self.active = True
		self.kind = kind
		self.vertices = (agent1, agent2)
		self.weight = weight
		self.directed = False
		if direction is not None:
			self.directed = True
			if isinstance(direction, int):
				if direction==0: self.directed = False
				elif direction>0: self.startpoint, self.endpoint = (agent1, agent2)
				elif direction<0: self.startpoint, self.endpoint = (agent2, agent1)
			elif isinstance(direction, bool):
				self.directed = direction
				if direction: self.startpoint, self.endpoint = (agent1, agent2)
			elif isinstance(direction, baseAgent):
				if direction not in self.vertices: raise ValueError(ï('Direction must select one of the agents as an endpoint.'))
				self.endpoint = direction
				self.startpoint = agent1 if direction==agent2 else agent2
			else: raise ValueError(ï('Direction must be either int, bool, or agent.'))
		if not self.directed:
			self.endpoint, self.startpoint, self.directed = (None, None, False)

		#Add object to each agent, and to the model
		for agent in self.vertices:
			if not kind in agent.edges: agent.edges[kind] = []
			if not self in agent.edges[kind]: agent.edges[kind].append(self) #Don't add self-links twice

		agent1.model.doHooks('edgeInit', [self, kind, agent1, agent2])

	def __repr__(self):
		pair, arrow = (self.vertices, '—') if not self.directed else ((self.startpoint, self.endpoint), '→')
		return f'<{self.__class__.__name__}: {pair[0].__class__.__name__} {pair[0].id} {arrow} {pair[1].__class__.__name__} {pair[1].id}>'

	def cut(self):
		for agent in self.vertices:
			if self in agent.edges[self.kind]: agent.edges[self.kind].remove(self) #Remove from agents
		self.active = False
		self.vertices[0].model.doHooks('edgeCut', [self])

	def partner(self, agent):
		if agent==self.vertices[0]: return self.vertices[1]
		elif agent==self.vertices[1]: return self.vertices[0]
		else: raise ValueError(ï('Agent {} is not connected to this edge.').format(agent.id))

	def reassign(self, oldagent, newagent):
		self.vertices = (self.partner(oldagent), newagent)
		oldagent.edges[self.kind].remove(self)
		newagent.edges[self.kind].append(self)
		newagent.model.doHooks('edgeReassign', [self, oldagent, newagent])

class Stocks:
	def __init__(self, breed, goodslist):
		self.goods = {g:{} for g in goodslist}
		for good, ginfo in goodslist.items():
			for p, fn in ginfo.props.items():
				endow = fn(breed) if callable(fn) else fn
				if endow is None: self.goods[good][p] = 0
				elif isinstance(endow, (tuple, list)): self.goods[good][p] = randint(*endow)
				else: self.goods[good][p] = endow

	def __getitem__(self, key):
		if isinstance(key, str): return self.goods[key]['quantity']
		elif isinstance(key, tuple):
			if isinstance(key[1], str): return self.goods[key[0]][key[1]]
			elif key[1] is True: return self.goods[key[0]]
			elif key[1] is False: return self.goods[key]['quantity']
		raise KeyError

	def __setitem__(self, key, val):
		if isinstance(key, str): self.goods[key]['quantity'] = val
		elif isinstance(key, tuple) and isinstance(key[1], str): self.goods[key[0]][key[1]] = val
		else: raise KeyError

	def __iter__(self): return iter({k: g['quantity'] for k,g in self.goods.items()})
	def __next__(self): return next({k: g['quantity'] for k,g in self.goods.items()})
	def __len__(self): return len(self.goods)
	def keys(self): return self.goods.keys()
	def values(self): return [g['quantity'] for g in self.goods.values()]
	def items(self): return [(k, g['quantity']) for k,g in self.goods.items()]


# ==========
# Helper classes and functions used in multiple modules
# None of these are intended to be user-facing
# ==========

import warnings
from numbers import Number
from sys import __stdout__
from io import BufferedWriter

#Internationalization since using _ is a disaster
#Can't install to global scope because it conflicts with readline;
#Can't name it _ here because `import *` skips it (???)
def ï(text): return helipad_gettext(text)

#Checks for any Ipython environment, including Spyder, for event loop purposes.
def isIpy():
	try:
		__IPYTHON__
		return True
	except NameError: return False

#Check whether Helipad is running in an interactive notebook. However, get_ipython() comes
#back undefined inside callbacks. So cache the value once, the first time it runs.
def isNotebook():
	if not '__helipad_ipy' in globals():
		try:
			globals()['__helipad_ipy'] = 'InteractiveShell' in get_ipython().__class__.__name__
		except NameError: globals()['__helipad_ipy'] = False

	return __helipad_ipy

def isBuffered():
	return isinstance(__stdout__.buffer, BufferedWriter)

#Generic extensible item class to store structured data
class Item:
	def __init__(self, **kwargs):
		for k,v in kwargs.items():
			setattr(self, k, v)

class funcStore(dict):
	multi = False

	def add(self, name, function):
		if self.multi:
			if name not in self: self[name] = [function]
			else: self[name].append(function)
		else: self[name] = function
		return function

	def remove(self, name, fname=None, removeall=False):
		if isinstance(name, (list, tuple)): return [self.remove(n) for n in name]
		if name not in self: return False
		if self.multi and fname:
			if isinstance(fname, (list, tuple)): return [self.remove(name, f, removeall) for f in fname]
			did = False
			for h in self[name]:
				if h.__name__ == fname:
					self[name].remove(h)
					if not removeall: return True
					else: did = True
			return did
		else:
			del self[name]
			return True

import colorsys, matplotlib.colors as mplcolor
class Color:
	def __init__(self, color):
		#Can take a hex string, color name, or [r,g,b] list/tuple.
		self.rgb = mplcolor.hex2color(color) if isinstance(color, str) else list(color)

	def __repr__(self): return f'<Color: ({round(self.r*100)},{round(self.g*100)},{round(self.b*100)})>'

	@property
	def hex(self): return mplcolor.to_hex(self.rgb)
	@property
	def hsv(self): return list(mplcolor.rgb_to_hsv(self.rgb))
	@property
	def r(self): return self.rgb[0]
	@property
	def g(self): return self.rgb[1]
	@property
	def b(self): return self.rgb[2]
	@property
	def h(self): return self.hsv[0]
	@property
	def s(self): return self.hsv[1]
	@property
	def v(self): return self.hsv[2]

	def lighten(self, factor=3):
		hls = colorsys.rgb_to_hls(*self.rgb)
		return Color(colorsys.hls_to_rgb(hls[0], (1-1/factor) + hls[1]/factor, hls[2]))

	def darken(self):
		hls = colorsys.rgb_to_hls(*self.rgb)
		return Color(colorsys.hls_to_rgb(hls[0], hls[1]-0.075 if hls[1]>0.075 else 0, hls[2]))

	def blend(self, color2):
		return Color(((self.r+color2.r)/2, (self.g+color2.g)/2, (self.b+color2.b)/2))

def makeDivisible(n, div, c='min'):
	return n-n%div if c=='min' else n+(div-n%div if n%div!=0 else 0)